/*
 * msmtp.c
 *
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@web.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>

#include "merror.h"
#include "tls.h"
#include "net.h"
#include "esmtp.h"

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
#include "getopt.h"
#endif /* HAVE_GETOPT_LONG */

/* not defined in smtp.h, because it is only used here: */
#define AUTH_METHOD_CHOOSE_SECURE	128

/* The name of this program */
char *prgname;
/* The version number */
#define VERSION "0.3.0"


/*
 * The main function processes the options.
 * It returns values from sysexits.h (like sendmail does).
 */

int main(int argc, char *argv[])
{
    /* the configuration */
    char *server = NULL;
    int  port = 25;
    char *from = NULL;
    char *auth_method_string = NULL;
    int auth_method = -1;
    char *user = NULL;
    char *password = NULL;
    char *tls_method_string = NULL;
    int tls_method = -1;
    int keep_bcc_flag = -1;
    /* the recipients of the mail */
    int rcptc;
    char **rcptv;
    /* misc */
    int error_code;
    /* commandline processing with getopt_long */
    int c;
    int option_index;
    struct option long_options[] =
    {
	{ "server",   required_argument, 0, 's' },
	{ "port",     required_argument, 0, 'P' },
	{ "from",     required_argument, 0, 'f' },
	{ "auth",     optional_argument, 0, 'a' },
	{ "user",     required_argument, 0, 'u' },
	{ "password", required_argument, 0, 'p' },
	{ "tls",      optional_argument, 0, 't' },
	{ "keep-bcc", no_argument,       0, 'l' },
	{ 0, 0, 0, 0 }
    };


    /* process the command line */
    prgname = strrchr(argv[0], '/');
    if (!prgname)
    {
	prgname = argv[0];
    }
    else
    {
	prgname++;
    }
    if (argc == 2)
    {
	if (strcmp(argv[1], "--version") == 0)
	{
	    printf("This is %s version %s\n", prgname, VERSION);
	    return 0;
	}
	else if (strcmp(argv[1], "--help") == 0)
	{
	    printf("Usage: %s <options>\n"
		    "Options are:\n"
		    "--server=smtpserver  set SMTP server\n"
		    "[--port=smtpport]    set port number\n"
		    "--from=addr          set your mail address\n"
		    "[--auth[=method]]    activate authentication\n"
		    "                     (and optionally choose method: "
		    "plain, login or cram_md5)\n"
		    "[--user=login]       set your username for "
		    "authentication\n"
		    "[--password=secret]  set your password for "
		    "authentication\n"
		    "[--tls[=method]]     activate TLS encryption\n"
		    "                     (and optionally choose method: "
		    "tls1, ssl3 or ssl2)\n"
		    "[--keep-bcc]         keep the Bcc header\n",
		    prgname);
	    return 0;
	}
    }
    error_code = 0;
    for (;;)
    {
	c = getopt_long(argc, argv, "s:P:f:a::u:p:t::l",
		long_options, &option_index);
	if (c == -1)
	{
	    break;
	}
	switch(c)
	{
	    case 's':
		server = optarg;
		break;

	    case 'P':
		port = atoi(optarg);
		break;

	    case 'f':
		from = optarg;
		break;

	    case 'a':
	    	auth_method_string = optarg;
    		auth_method = AUTH_METHOD_CHOOSE_SECURE;
		break;
		
	    case 'u':
		user = optarg;
		break;

	    case 'p':
		password = optarg;
		break;

	    case 't':
	    	tls_method_string = optarg;
    		tls_method = TLS_METHOD_CHOOSE;
		break;

	    case 'l':
		keep_bcc_flag = 1;
		break;
	    
	    default:
		error_code = 1;
		break;
	}
	if (error_code)
	{
	    break;
	}
    }
    if (!error_code)
    {
	if (!server)
	{
	    add_error_line("need to know the smtp server: use --server");
	    error_code = 1;
	}
	if (port < 1 || port > 65535)
	{
	    add_error_line("invalid port number %d", port);
	    error_code = 1;
	}
	if (!from)
	{
	    add_error_line("need to know your mail address: use --from");
	    error_code = 1;
	}
	if (tls_method_string)
	{
	    if (strcmp(tls_method_string, "tls1") == 0)
	    {
		tls_method = TLS_METHOD_TLS1;
	    }
	    else if (strcmp(tls_method_string, "ssl3") == 0)
	    {
		tls_method = TLS_METHOD_SSL3;
	    }
	    else if (strcmp(tls_method_string, "ssl2") == 0)
	    {
		tls_method = TLS_METHOD_SSL2;
	    }
	    else
	    {
		add_error_line("unknown TLS method %s", tls_method_string);
		error_code = 1;
	    }
	}
	if (tls_method == -1)
	{
	    tls_method = TLS_METHOD_NONE;
	}
	if (auth_method_string)
	{
	    if (strcmp(auth_method_string, "plain") == 0)
	    {
		auth_method = AUTH_METHOD_PLAIN;
	    }
	    else if (strcmp(auth_method_string, "login") == 0)
	    {
		auth_method = AUTH_METHOD_LOGIN;
	    }
	    else if (strcmp(auth_method_string, "cram_md5") == 0)
	    {
		auth_method = AUTH_METHOD_CRAM_MD5;
	    }
	    else
	    {
		add_error_line("unknown SMTP AUTH method %s", auth_method_string);
		error_code = 1;
	    }
	}
	if (auth_method == -1 && (user || password))
	{
	    auth_method = AUTH_METHOD_CHOOSE_SECURE;
	}
	if (auth_method == -1)
	{
	    auth_method = AUTH_METHOD_NONE;
	}
	if (auth_method != AUTH_METHOD_NONE && (!user || !password))
	{
	    add_error_line("need both username and password for "
		    "SMTP authentication");
	    error_code = 1;
	}
	if (auth_method == AUTH_METHOD_CHOOSE_SECURE)
	{
	    /* don't send cleartext passwords! */
	    if (tls_method == TLS_METHOD_NONE)
	    {
		auth_method = AUTH_METHOD_CRAM_MD5;
	    }
	    else
	    {
		auth_method = AUTH_METHOD_PLAIN;
	    }
	}
	if (keep_bcc_flag == -1)
	{
	    keep_bcc_flag = 0;
	}
    }
    if (!error_code)
    {
	rcptc = argc - optind;
	rcptv = &(argv[optind]);
	if (rcptc == 0)
	{
	    add_error_line("no recipients given");
	    error_code = 1;
	}
    }
    if (error_code)
    {
	print_error_lines(stderr);
	return EX_USAGE;
    }

    /* DEBUG
    printf("server = %s\nport = %d\nfrom = %s\n"
	    "auth = %d\nuser = %s\npw = %s\n"
	    "tls = %d\n"
	    "keep-bcc = %d\n",
	    server, port, from,
	    auth_method, user, password, 
	    tls_method,
	    keep_bcc_flag);
    */

    if ((error_code = smtp_send(server, port, from,
		    auth_method, user, password, 
		    tls_method, keep_bcc_flag,
		    rcptc, rcptv, 
		    stdin)) != 0)
    {
	add_error_line("could not send mail");
	print_error_lines(stderr);
	return error_code;
    }
	
    return 0;
}
