/*
 * net.c
 * 
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@web.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <string.h>
#include <errno.h>
extern int errno;
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "merror.h"
#include "tls.h"
#include "net.h"
#include "xmalloc.h"


/*
 * open_socket() 
 *
 * Opens a TCP socket to 'hostname':'port'.
 * IPv6 will be used if possible, otherwise IPv4.
 * On success, 0 is returned and the file descriptor is stored in 'fd'. 
 * On failure, 1 is returned and the value of 'fd' is unspecified.
 */

int open_socket(char *hostname, int port, int *fd)
{
    char *port_string;
    struct addrinfo hints;
    struct addrinfo *res0;
    struct addrinfo *res;
    int error_code;
    int cause;
    
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = 0;
    hints.ai_protocol = 0;
    hints.ai_addrlen = 0;
    hints.ai_canonname = NULL;
    hints.ai_addr = NULL;
    hints.ai_next = NULL;
    xasprintf(&port_string, "%d", port);
    error_code = getaddrinfo(hostname, port_string, &hints, &res0);
    free(port_string);
    if (error_code)
    {
	add_error_line("cannot locate host %s [%s]",
		hostname, gai_strerror(error_code));
	return 1;
    }

    *fd = -1;
    cause = 0;
    for (res = res0; res; res = res->ai_next)
    {
	*fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (*fd < 0)
	{
	    cause = 1;
	    continue;
	}
	if (connect(*fd, res->ai_addr, res->ai_addrlen) < 0) 
	{
	    cause = 2;
	    close(*fd);
	    *fd = -1;
	    continue;
	}
	break;
    }
    freeaddrinfo(res0);
    if (*fd < 0)
    {
	if (cause == 1)
	{
	    add_error_line("cannot create socket [socket(): %s]", 
		    strerror(errno));
	}
	else /* cause == 2 */
	{
	    add_error_line("cannot connect to %s, port %d [connect(): %s]", 
		    hostname, port, strerror(errno));
	}
	return 1;
    }
    
    return 0;
}


/*
 * net_getchar()
 *
 * Reads a character from 'fd' and stores it in 'c'.
 * If 'ssl' is not NULL, the read operation uses TLS. 
 * Return value is 0 on success, -1 on EOF and 1 on error.
 */

int net_getchar(int fd, SSL *ssl, char *c)
{
    ssize_t error_code;
    
    if (ssl)
    {
	return tls_getchar(ssl, c);
    }
    else
    {
	error_code = read(fd, c, (size_t)1);
	if (error_code != 1)
	{
	    add_error_line("read error [read(): %s]", strerror(errno));
	    return (error_code == 0) ? -1 : 1;
	}
	return 0;
    }
}


/*
 * net_getline()
 * 
 * Reads a line from 'fd' into 'line', which is of size 'size'.
 * If 'ssl' is not NULL, the read operation uses TLS.
 * '\r's are ignored. The ending '\n' is not stored in 'line'. If the line is
 * longer than 'size - 1' characters, only these characters will be stored in
 * 'line'. Reading can start again at the next character of the line. The
 * line will always be terminated with '\0'. 
 * Return value is 0 on success and 1 on read error.
 */

int net_getline(int fd, SSL *ssl, char *line, int size)
{
    char c;
    int i;
    int error_code;

    i = 0;
    while ((error_code = net_getchar(fd, ssl, &c)) == 0)
    {
	if (c == '\n')
	{
	    break;
	}
	else if (c == '\r')
	{
	    /* ignore '\r's */
	}
	else
	{
	    line[i++] = c;
	    if (i == size - 1)
	    {
		break;
	    }
	}
    }
    line[i] = '\0';
    if (error_code == 0 || error_code == -1)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}


/*
 * net_putline()
 *
 * Writes a line to 'fd'.
 * If 'ssl' is not NULL, the write operation uses TLS.
 * Return value is 0 on success and 1 on write error.
 */

int net_putline(int fd, SSL *ssl, char *line)
{
    ssize_t error_code;
    size_t count;

    if (ssl)
    {
	return tls_putline(ssl, line);
    }
    else
    {
	count = strlen(line);    
	error_code = write(fd, line, count);
	if (error_code < count)
	{
	    add_error_line("write error [write(): %s]", strerror(errno));
	    return 1;
	}
	else
	{
	    return 0;
	}
    }
}
