/*
 * tls.c
 *
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@web.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <limits.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>

#include "merror.h"
#include "tls.h"

#define SSL_ERROR_BUF 120	/* 120 is enough according to 
				   ERR_error_string(3ssl) */

/*
 * init_tls()
 *
 * Initialize SSL library and an SSL_CTX for a client connection.
 * 'method' must be one of the methods defined in tls.h
 * Returns SSL_CTX on success and NULL on error.
 */

SSL_CTX *tls_init(int method)
{
    SSL_CTX *ssl_ctx = NULL;
    SSL_METHOD *ssl_method = NULL;
    char ssl_error_string[SSL_ERROR_BUF];
    
    SSL_load_error_strings();
    SSL_library_init();
    if (!RAND_status)
    {
	/* FIXME: seed random number generator 
	 * On systems whith /dev/urandom, OpenSSL seeds itself.
	 * On other systems, we have to do it ourselves with RAND_seed().
	 * I have no idea with what. */
	ERR_error_string(ERR_get_error(), ssl_error_string);
	add_error_line("BUG: must seed random number generator for TLS [%s]", 
		ssl_error_string);
	return NULL;
    }
    switch (method)
    {
	case TLS_METHOD_CHOOSE:
	    ssl_method = SSLv23_client_method();
	    break;

	case TLS_METHOD_SSL2:
	    ssl_method = SSLv2_client_method();
	    break;
	    
	case TLS_METHOD_SSL3:
	    ssl_method = SSLv3_client_method();
	    break;

	case TLS_METHOD_TLS1:
	    ssl_method = TLSv1_client_method();
	    break;

	default:
	    add_error_line("BUG: unknown TLS method in init_tls()");
	    return NULL;
    }
    if (!ssl_method)
    {
	add_error_line("TLS method %d failed", method);
	return NULL;
    }
    ssl_ctx = SSL_CTX_new(ssl_method);
    if (!ssl_ctx)
    {
	ERR_error_string(ERR_get_error(), ssl_error_string);
	add_error_line("cannot activate TLS support [%s]", ssl_error_string);
	return NULL;
    }

    return ssl_ctx;
}


/*
 * start_tls
 *
 * Starts TLS encryption on a socket.
 * Returns SSL* on success and NULL on error.
 */

SSL *tls_start(SSL_CTX *ssl_ctx, int fd)
{
    SSL *ssl;
    int error_code;
    unsigned long error_code_2;
    char ssl_error_string[SSL_ERROR_BUF];
    
    if (!(ssl = SSL_new(ssl_ctx)))
    {
	ERR_error_string(ERR_get_error(), ssl_error_string);
	add_error_line("SSL_new() failed [%s]", ssl_error_string);
	return NULL;
    }
    if (!SSL_set_fd(ssl, fd))
    {
	ERR_error_string(ERR_get_error(), ssl_error_string);
	add_error_line("SSL_set_fd() failed [%s]", ssl_error_string);
	SSL_free(ssl);
	return NULL;
    }
    if ((error_code = SSL_connect(ssl)) < 1)
    {
	error_code = SSL_get_error(ssl, error_code);
	switch (error_code)
	{
	    case SSL_ERROR_ZERO_RETURN:
		add_error_line("cannot establish TLS connection "
			"[TLS connection was closed unexpectedly]");
		break;

	    case SSL_ERROR_SYSCALL:
		error_code_2 = ERR_get_error();
		if (error_code_2 == 0)
		{
		    if (error_code == 0)
		    {
			add_error_line("cannot establish TLS connection "
				"[a protocol violating EOF occured in "
				"SSL_connect()]");
			break;
		    }
		    else if (error_code == -1)
		    {
			add_error_line("cannot establish TLS connection "
				"[input/output error: %s]", strerror(errno));
		    }
		}
		else
		{
		    add_error_line("cannot establish TLS connection [%s]", 
			    ssl_error_string);
		}
		break;

	    case SSL_ERROR_SSL:
		ERR_error_string(ERR_get_error(), ssl_error_string);
		add_error_line("cannot establish TLS connection [%s]", 
			ssl_error_string);
		break;

	    default:
		add_error_line("cannot establish TLS connection "
			"[unknown error]");
	}
	SSL_free(ssl);
	return NULL;
    }
    if (!SSL_get_peer_certificate(ssl))
    {
	ERR_error_string(ERR_get_error(), ssl_error_string);
	add_error_line("cannot get server certificate [%s]", ssl_error_string);
	SSL_free(ssl);
	return NULL;
    }
    /* TODO: Do something with the server certificate */

    return ssl;
}


/*
 * tls_getchar()
 *
 * Reads a character using TLS and stores it in 'c'.
 * Returns 0 on success, -1 on EOF and 1 on error.
 */

int tls_getchar(SSL *ssl, char *c)
{
    unsigned long error_code;
    char ssl_error_string[SSL_ERROR_BUF];
    
    if (SSL_read(ssl, c, 1) < 1)
    {
	/* TODO: error handling could be better */
	error_code = ERR_get_error();
	if (!error_code)
	{
	    /* EOF */
	    return -1;
	}
	else
	{
	    ERR_error_string(error_code, ssl_error_string);
	    add_error_line("cannot read from TLS connection [%s]", 
		    ssl_error_string);
	    return 1;
	}
    }
    else
    {
	return 0;
    }
}


/*
 * tls_putline()
 *
 * Writes a line using TLS.
 * Returns 1 on error and 0 on success.
 */

int tls_putline(SSL *ssl, char *line)
{
    size_t count;
    unsigned long error_code;
    char ssl_error_string[SSL_ERROR_BUF];

    count = strlen(line);
    if (count > (size_t)INT_MAX)
    {
	add_error_line("line too long in tls_putline()");
	return 1;
    }
    if (SSL_write(ssl, line, (int)count) != (int)count)
    {
	/* TODO: error handling could be better */
	error_code = ERR_get_error();
	if (!error_code)
	{
	    add_error_line("cannot write to TLS connection [unknown error]");
	}
	else
	{
	    ERR_error_string(error_code, ssl_error_string);
	    add_error_line("cannot write to TLS connection [%s]", 
		    ssl_error_string);
	}
	return 1;	
    }

    return 0;
}


/*
 * tls_close()
 *
 * Close 'ssl', free 'ssl' and 'ssl_ctx'
 */

void tls_close(SSL *ssl, SSL_CTX *ssl_ctx)
{
    if (ssl)
    {
	SSL_shutdown(ssl);
	SSL_free(ssl);
    }
    if (ssl_ctx)
    {
	SSL_CTX_free(ssl_ctx);
    }
}
